const express = require('express');
var url = require('url');
var fs = require('fs');

var portnumber = 3001;
const server = express();
server.use(express.static("."));
server.listen(portnumber);
console.log("Now listening on port " + portnumber);

server.get('/', sendIndex);
server.get('/index.html', sendIndex);
server.get('/click', handleClick);
server.get('/change-modus', handleChangeModus)



const WIDTH = 24;
const HEIGHT = 12;

const BOMB_PROB = 0.1;

let minefield = [];
let gameOver = false;
let gameWon = false;

MARK_BOMBS = 0;
CLICK_BOMBS = 1;

let modus = CLICK_BOMBS;


UNMARKED = 0;
MARKED = 1;
TENTATIVELY_MARKED = 2;
MARKED_MODUSES = 3;

/*
   field = struct {
    i : Int;
    j : Int;
    bomb : Bool;
    uncovered : Bool;
    marked : Int;
    bombNeighbors : Int;
   }
*/

function newMineField(width, height, bombProb) {
    let theMinefield = [];
    for (let i = 0; i < height; i++) {
        theMinefield[i] = [];
        for (let j = 0; j < width; j++) {
            let field = {};
            field.i = i;
            field.j = j;
            field.bomb = (Math.random() < bombProb);
            field.uncovered = false;
            field.marked = UNMARKED;
            field.bombNeighbors = 0;
            theMinefield[i][j] = field;
        }
    }
    function isBomb(y, x) {
        if (y < 0 || y >= HEIGHT || x < 0 || x >= WIDTH)
            return false;
        else
            return theMinefield[y][x].bomb;
    }
    for (let i = 0; i < height; i++) {
        for (let j = 0; j < width; j++) {
            let field = theMinefield[i][j];
            for (let di = -1; di <= 1; di++) {
                for (let dj = -1; dj <= 1; dj++) {
                    if (isBomb(i + di, j + dj)) {
                        field.bombNeighbors++;
                    }
                }
            }
        }
    }
    return theMinefield;
}

function fieldToTD(field) {
    let fieldname = '';
    if (field.uncovered == false) {
        if (field.marked == MARKED) {
            filename = 'flag.svg';
        }
        else if (field.marked == TENTATIVELY_MARKED) {
            filename = 'flag-tentative.svg';
        }
        else {
            filename = 'covered.svg';
        }
    }
    else if (field.bomb) {
        filename = 'bomb.svg';
    }
    else {
        filename = `template-0${field.bombNeighbors}-01.svg`;
    }
    if (gameOver || gameWon)
        return `<td><img src="img/${filename}"></td>`;
    else
        return `<td><a href="click?x=${field.j}&y=${field.i}"><img src="img/${filename}"></a></td>`;
}
function minefieldToTableString(minefield) {
    let tablerows = '';
    for (let j = 0; j < HEIGHT; j++) {
        let trInnerString = ''
        for (let i = 0; i < WIDTH; i++) {
            trInnerString += fieldToTD(minefield[j][i]) + '\n';
        }
        tablerows += `<tr>${trInnerString}</tr>\n`;
    }
    return `<table cellspacing="0"><tbody>${tablerows}</tbody></table></html>`;
}


function minefieldHtmlString(minefield) {
    let gameOverString = '';
    let gameWonString = '';
    let markOrClickString = markOrClickDiv();
    if (gameOver) {
        gameOverString = "\n<a href='index.html'><div class='gameover'>Game Over</div></a>\n";
    }
    if (gameWon) {
        gameWonString = "\n<a href='index.html'><div class='gamewon'>Game Won</div></a>\n";
    }
    let minefieldString = minefieldToTableString(minefield);
    let htmlString =
        `<!DOCTYPE html>
    <html lang="en">
    <head>
      <meta name="viewport" content="initial-scale=1.0, width=device-width" />
      <title>Minesweeper</title>
      <meta charset="utf-8">
      <link rel="stylesheet" type="text/css" href="minesweeper.css" />
    </head>

    <body>
    <div class='topdiv'>
    ${markOrClickString}
    ${minefieldString}
    ${gameOverString}
    ${gameWonString}
    </div>
    </body>
    </html>
    `;
    return htmlString;
}


function markOrClickDiv() {
    let result = '';
    if (modus == CLICK_BOMBS) {
        result =
            `
        <div class='mark-or-click'>
            <a href='change-modus'>mark</a> 
            <strong>click</strong>
        </div>
        `;
    }
    else {
        result =
            `
        <div class='mark-or-click'>
        <strong>mark</strong>
        <a href='change-modus'>click</a>            
        </div>
        `;
    }

    return result;
}



function sendIndex(req, res) {
    console.log('sendIndex called');
    minefield = newMineField(WIDTH, HEIGHT, BOMB_PROB);
    gameOver = false;
    gameWon = false;
    modus = CLICK_BOMBS;
    res.send(minefieldHtmlString(minefield));
}


function uncover(x, y) {
    console.log(`uncover(${x}, ${y}) called;`);
    if (x < 0 || x >= WIDTH || y < 0 || y >= HEIGHT)
        return;
    let field = minefield[y][x];
    console.log(`minefield here is ${JSON.stringify(field)}`);
    if (minefield[y][x].uncovered)
        return;
    minefield[y][x].uncovered = true;
    if (minefield[y][x].bombNeighbors > 0)
        return;

    uncover(x - 1, y);
    uncover(x + 1, y);
    uncover(x, y - 1);
    uncover(x, y + 1);
    uncover(x - 1, y - 1);
    uncover(x - 1, y + 1);
    uncover(x + 1, y - 1);
    uncover(x + 1, y + 1);
}


function everythingUncovered(minefield) {
    for (let j = 0; j < HEIGHT; j++) {
        for (let i = 0; i < WIDTH; i++) {
            if (minefield[j][i].uncovered == false) {
                if (minefield[j][i].marked == TENTATIVELY_MARKED) {
                    return false;
                }
                if (minefield[j][i].marked == UNMARKED) {
                    console.log(`everythingUncovered false because (${i},${j}) not marked`);
                    return false;
                }
                if (minefield[j][i].bomb == false) {
                    console.log(`everythingUncovered false because (${i},${j}) has unmarked bomb`);
                    return false;
                }
            }
        }
    }
    console.log('everythingUncovered returns true');
    return true;
}

function handleClick(req, res) {
    let x = parseInt(req.query.x);
    let y = parseInt(req.query.y);
    console.log(`point (${x},${y}) ${modus == CLICK_BOMBS ? 'clicked' : 'marked'}`);
    console.log(`modus is ${modus == CLICK_BOMBS ? "CLICK_BOMS" : "MARK_BOMBS"}`);

    if (modus == MARK_BOMBS || modus == TENTATIVELY_MARKED) {
        if (minefield[y][x].uncovered == false) {
            minefield[y][x].marked = (minefield[y][x].marked + 1) % MARKED_MODUSES;
        }
        else {
            // Do nothing when trying to mark/unmark an uncovered field!
        }
    }
    else { // if modus == CLICK_BOMBS
        if (minefield[y][x].marked != UNMARKED) {
            // do nothing if user tries to click a  marked field or a tentatively marked field
        }
        else if (minefield[y][x].bomb) {
            console.log(`game over`);
            gameOver = true;
            for (let j = 0; j < HEIGHT; j++) {
                for (let i = 0; i < WIDTH; i++) {
                    minefield[j][i].uncovered = true;
                }
            }
        }
        else {
            uncover(x, y);
        }
    }
    if (everythingUncovered(minefield) && !gameOver) {
        gameWon = true;
    }
    let mineFieldString = minefieldHtmlString(minefield);
    res.send(mineFieldString);
}

function handleChangeModus(req, res) {
    console.log("handleChangeModus called");
    if (modus == MARK_BOMBS) {
        modus = CLICK_BOMBS;
    }
    else {
        modus = MARK_BOMBS;
    }
    let mineFieldString = minefieldHtmlString(minefield);
    res.send(mineFieldString);
}